#ifndef IVCSERVER__H
#define IVCSERVER__H

#include <string>
#include <exception>
#include <ostream>

#include <QMap>
#include <QObject>
#include <QLocalSocket>

#include <xen/be/Exception.hpp>
#include <xen/be/Log.hpp>

extern "C" {
#include <libivc.h>
};

#include "libivc_core.h"

// Forward declaration.
class IvcBackend;

//
// Exception generated by IvcServer.
//
class IvcException : public XenBackend::Exception {
    using Exception::Exception;
};

//
// Keep and organize IVC connection data.
//
class IvcConnection {
public:
    IvcConnection(domid_t domid, domid_t peerDomid, uint16_t port,
            uint64_t id, uint32_t eventChannel);
    IvcConnection(const libivc_message_t& message);
    ~IvcConnection();
    bool operator<(const IvcConnection& other) const;

    // Accessors.
    uint16_t getDomid() const { return mDomid; };
    uint16_t getPeerDomid() const { return mPeerDomid; };
    uint16_t getPort() const { return mPort; };
    uint64_t getId() const { return mId; };
    uint64_t getEventChannel() const { return mEventChannel; };

    IvcConnection reverse() const {
        return IvcConnection(mPeerDomid, mDomid, mPort, mId, mEventChannel);
    }
    std::ostream& print(std::ostream& os) const;
    libivc_message_t makeControlMessage(MESSAGE_TYPE_T type, int16_t status) const;

private:
    domid_t mDomid;
    domid_t mPeerDomid;
    uint16_t mPort;
    uint64_t mId;
    uint32_t mEventChannel;
};

//
// IVC server.
// Keeps track of IVC Stream connection states established between a backend
// server and IVC frontends.
// IVC-daemon is the IVC backend, as it holds the IVC control-ring over which
// IVC stream connections are established after CONNECT->ACK dance.
// Should a backend server vanish/crash, this object will notify that the
// stream is disconnected.
//
class IvcServer : public QObject {
    Q_OBJECT
public:
    IvcServer(IvcBackend& be, QLocalSocket* mSock);
    virtual ~IvcServer();

public slots:
    // Slot for localDomain to other domain communication.
    void onEgressMessage();
    // Slot for other domain to local domain communication.
    void onIngressMessage(const libivc_message_t& msg);
    void flushDomain(domid_t domid);

private:
    IvcServer(const IvcServer&) = delete;
    IvcServer& operator=(const IvcServer&) = delete;

    int processMessage(const libivc_message_t& msg,
            QMap<IvcConnection, MESSAGE_TYPE_T>& local,
            QMap<IvcConnection, MESSAGE_TYPE_T>& remote);
    int processMessage(const libivc_message_t& msg);
    void disconnectAll();
    void flushConnections(
            QMap<IvcConnection, MESSAGE_TYPE_T>& connections,
            domid_t domid);

    // Maps of established connections and their states.
    // It is assumed this server will never send CONNECT and never initiate
    // connections.
    QMap<IvcConnection, MESSAGE_TYPE_T> mConnections;
    // Established connection with the server.
    // (result of QLocalServer::nextPendingConnection()).
    QLocalSocket* mSock;
    // Reference to backend owner for IVC communication.
    IvcBackend& mBackend;

    XenBackend::Log mLog;
};

inline std::ostream& operator<<(std::ostream& os, const IvcConnection& con)
{ return con.print(os); }

#endif /* IVCSERVER__H */
